// BUNAK revision 2, the non-synthetic RAM bandwidth reporter, it sums a given memory block with granularity 64bits i.e. 8bytes.
// Written by ChatGPT 3.5 with help of Kaze (sanmayce@sanmayce.com), hee-hee.
// How to compile: $ gcc -O3 -o BuBa_BunakBandwidth_YMM.elf -fopenmp BuBa_BunakBandwidth_YMM.c -mavx2 -D_N_HIGH_PRIORITY_GCC
// How to compile: $ x86_64-w64-mingw32-gcc -O3 -m64 -o BuBa_BunakBandwidth_YMM.exe -fopenmp BuBa_BunakBandwidth_YMM.c -mavx2 -D_N_HIGH_PRIORITY
// How to run: $ /bin/time -v ./BuBa_BunakBandwidth_YMM.elf
// With above run you can be sure the memory footprint to be 0.5GB*8, this is done for machinettes with 8GB.

#include <stdio.h>
#include <stdlib.h>
//#include <x86intrin.h> // Include Intel Intrinsics
#include <immintrin.h> // Include AVX2 Intrinsics
#include <omp.h>
#include <time.h> // Include for time()
#include <string.h> // strcmp()

//#define _WIN32_ENVIRONMENT_
//#define _POSIX_ENVIRONMENT_

#ifdef _N_HIGH_PRIORITY
// https://msdn.microsoft.com/en-us/library/windows/desktop/ms686219.aspx
//#include <stdio.h>
#include <windows.h>
#include <tchar.h>
#endif

#ifdef _N_HIGH_PRIORITY_GCC
//#include <sys/time.h>
#include <sys/resource.h>
//int main(){
//    setpriority(PRIO_PROCESS, 0, -20);
//}
//Note that you must be running as superuser for this to work. (for more info, type 'man setpriority' at a prompt.)
#endif

//#define ARRAY_SIZE (1 << 30) // 1GB array size
size_t ARRAY_SIZE = 1*512*1024*1024ULL; // 1x0.5GB * 8 in order to fit into 8 GB
#define NUM_RUNS 7

// Function to initialize the array with random values
void initializeArray(long long *array, size_t size) {
    printf("Initializing block %lld MB ...\n", (long long)(double)(size * sizeof(long long))>>20);
    // Caramba?! Why when commented is much faster?!
    //#pragma omp parallel for
    for (size_t i = 0; i < size; ++i) {
        array[i] = 13*rand();
        //if ( i%(1024*1024)==0 ) printf("%d%%\r", (int)(i*100/size));
    }
}

// Function to benchmark RAM read bandwidth
double benchmarkBandwidth(long long *array, size_t size) {
    // Initialize vectors
    __m128 vec;
    float sum_arr[4] = {0.0f, 0.0f, 0.0f, 0.0f};

    // Measure the start time
    double startMP = omp_get_wtime();
    time_t start, end;
    clock_t startCLCK, endCLCK;
    time(&start);
    startCLCK = clock();

//XMM:
/*
    // Vectorized memory read
    #pragma omp parallel for
    for (size_t i = 0; i < size; i += 4) {
        vec = _mm_loadu_ps(&array[i]);
        float temp[4];
        _mm_storeu_ps(temp, vec);
        #pragma omp simd
        for (int j = 0; j < 4; j++) {
            sum_arr[j] += temp[j];
        }
    }

    // Perform horizontal sum
    float total_sum = 0.0f;
    for (int j = 0; j < 4; j++) {
        total_sum += sum_arr[j];
    }
*/

    // Vectorized memory read
    long long sum = 0;
    #pragma omp parallel
    {
        long long local_sum = 0;
        #pragma omp for
        for (size_t i = 0; i < size; i += 4) {
            __m256i vecYMM = _mm256_loadu_si256((__m256i *)&array[i]);
            long long temp[4];
            _mm256_storeu_si256((__m256i *)temp, vecYMM);
            for (int j = 0; j < 4; j++) {
                local_sum += temp[j];
            }
        }
        #pragma omp atomic
        sum += local_sum;
    }

    // Measure the end time
    double endMP = omp_get_wtime();
    time(&end);
    endCLCK = clock();

    // Calculate bandwidth
    double timew = (endMP - startMP);
    double timeC = (end - start);
    double timeCLCK = ((double)(endCLCK - startCLCK)) / CLOCKS_PER_SEC;
    double bytes = (double)(size * sizeof(long long));
    double bandwidth = bytes / timew;

    printf("omp_get_wtime(): %.3f seconds\n", timew);
    //printf("C standard library time(): %.4f seconds\n", timeC);
    //printf("C standard library clock(): %.2f seconds\n", timeCLCK);

    return bandwidth;
}

int main(int argc, char **argv) {
    int MAXthreads;
    int MINthreads;

#ifdef _N_HIGH_PRIORITY
   DWORD dwError, dwPriClass;
#endif

printf("__________     __________                 __________                     __   __________                   .___       .__    .___ __  .__     \n");
printf("\\______   \\__ _\\______   \\____            \\______   \\__ __  ____ _____  |  | _\\______   \\____    ____    __| _/_  _  _|__| __| _//  |_|  |__  \n");
printf(" |    |  _/  |  \\    |  _|__  \\    ______  |    |  _/  |  \\/    \\\\__  \\ |  |/ /|    |  _|__  \\  /    \\  / __ |\\ \\/ \\/ /  |/ __ |\\   __\\  |  \\ \n");
printf(" |    |   \\  |  /    |   \\/ __ \\_ /_____/  |    |   \\  |  /   |  \\/ __ \\|    < |    |   \\/ __ \\|   |  \\/ /_/ | \\     /|  / /_/ | |  | |   Y  \\\n");
printf(" |______  /____/|______  (____  /          |______  /____/|___|  (____  /__|_ \\|______  (____  /___|  /\\____ |  \\/\\_/ |__\\____ | |__| |___|  /\n");
printf("        \\/             \\/     \\/                  \\/           \\/     \\/     \\/       \\/     \\/     \\/      \\/                \\/           \\/ \n");


printf("BuBa revision 2, the non-synthetic AVX2 vectorized OpenMP parallelized RAM bandwidth reporter, it sums a given memory block with granularity 64bits i.e. 8bytes.\n");
printf("Usage: BuBa [memory in GB] [max].\n");
#ifdef _N_HIGH_PRIORITY_GCC
setpriority(PRIO_PROCESS, 0, -20);

// setpriority(3) - Linux man page
// #include <sys/resource.h>
// int getpriority(int which, id_t who);
// int setpriority(int which, id_t who, int value); 
// The getpriority() function shall obtain the nice value of a process, process group, or user. The setpriority() function shall set the nice value of a process, process group, or user to value+ {NZERO}.
// Target processes are specified by the values of the which and who arguments. The which argument may be one of the following values: PRIO_PROCESS, PRIO_PGRP, or PRIO_USER, indicating that the who argument is to be interpreted as a process ID, a process group ID, or an effective user ID, respectively. A 0 value for the who argument specifies the current process, process group, or user.
//int which = PRIO_PROCESS;
//id_t pid;
//int ret;
//pid = getpid();
//ret = getpriority(which, pid);

printf("Current priority is %d.\n",getpriority(PRIO_PROCESS, 0));
#endif

#ifdef _N_HIGH_PRIORITY
   if(!SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS))
   {
//      _tprintf(TEXT("Already REALTIME_PRIORITY.\n"));
//      goto Cleanup;
   } 
   if(!SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS))
   {
//      _tprintf(TEXT("Already REALTIME_PRIORITY.\n"));
//      goto Cleanup;
   } 
   // Display priority class

   dwPriClass = GetPriorityClass(GetCurrentProcess());

   //_tprintf(TEXT("Current priority class is 0x%x\n"), dwPriClass);

//if (argc == 2) {} else // 2020-Jan-13
   if (dwPriClass==0x00000080) printf("Current priority class is HIGH_PRIORITY_CLASS.\n");
//if (argc == 2) {} else // 2020-Jan-13
   if (dwPriClass==0x00000100) printf("Current priority class is REALTIME_PRIORITY_CLASS.\n");
#endif

#ifdef _N_HIGH_PRIORITY
// https://habrahabr.ru/post/113682/
// Andrew Aksyonoff [

//SetProcessAffinityMask(GetCurrentProcess(), 1); // 2020-Dec-09, causes problems in real-time priority with newer Windows 10

    //volatile int zomg = 1;
    //for ( int i=1; i<1000000000; i++ )
    //  zomg *= i;
// Andrew Aksyonoff ]
#endif

    // Allocate memory for the array
if (argc>1) {
	ARRAY_SIZE=((long long)atoi(argv[1])*(1<<30ULL))>>3;
}
    long long *array = (long long *)malloc(ARRAY_SIZE * sizeof(long long));
    if (array == NULL) {
        printf("Memory allocation failed.\n");
        return 1;
    }

    // Initialize the array with random values
    initializeArray(array, ARRAY_SIZE);
    //printf("\nDone creating.\n");

    MAXthreads=omp_get_max_threads();
    MINthreads=1;//MAXthreads;

if (argc>=3) {
    if ( (strcmp(argv[2],"max")==0) || (strcmp(argv[2],"MAX")==0) ) MINthreads=MAXthreads;
}
	printf("Cycling through %d to %d threads...\n", MINthreads, MAXthreads);

    // Benchmark the RAM read bandwidth
    for (int num_threads = MINthreads; num_threads <= MAXthreads; ++num_threads) {
        omp_set_num_threads(num_threads);
        double max_bandwidth = 0.0;

        // Warm-up cache
/*
        for (int i = 0; i < 100; ++i) {
            double bandwidth = benchmarkBandwidth(array, ARRAY_SIZE);
            if (bandwidth > max_bandwidth) {
                max_bandwidth = bandwidth;
            }
        }
*/

        // Benchmark the RAM read bandwidth
        for (int run = 0; run < NUM_RUNS; ++run) {
            double bandwidth = benchmarkBandwidth(array, ARRAY_SIZE);
            if (bandwidth > max_bandwidth) {
                max_bandwidth = bandwidth;
            }
        }

        // Print the maximum observed bandwidth for the current num_threads scenario
        printf("Threads: %d, Maximum observed bandwidth: %.2f GB/s\n", num_threads, max_bandwidth / (1e9));
    }

    // Free allocated memory
    free(array);
    exit (0);
}


